/**
 * Grille
 * 
 * @package 	_Morpion
 * @author  	KOVAC Alexandre
 * @version 	1.0.0
 * @link 		<http://www.kovacou.fr/University/Project/Morpion>
 */

package _SRC;
import java.util.*;

public class Grille 
{
	// Constants
	final private int DEFAULT_SIZE = 5;
	
	// Proprits de la grille
	private int m_l;
	private int m_c;
	private Cellule[][] m_grille;
	
	// Historique des coups et des alignements
	private ArrayList<Cellule> m_journal;
	private ArrayList<Alignement> m_alignX;
	private ArrayList<Alignement> m_alignO;
	
	/**
	 * Constructeur de la grille.
	 * Construit une grille de L ligne et de C colonne en respectant les limitations imposes.
	 * 
	 * @param l Nombre de ligne
	 * @param c Nombre de colonne
	 * @throws Throwable
	 */
	public Grille(int l, int c) throws Throwable
	{
		if (l < DEFAULT_SIZE)
			throw new Throwable("-2.1");
		
		if (c < DEFAULT_SIZE)
			throw new Throwable("-2.2");
		
		this.m_l       = l;
		this.m_c       = c;
		this.m_grille  = new Cellule[l][c];
		this.m_journal = new ArrayList<Cellule>();
		this.m_alignO  = new ArrayList<Alignement>();
		this.m_alignX  = new ArrayList<Alignement>();
	}
	
	/**
	 * Retourner le nombre de ligne de la grille.
	 * 
	 * @return Nombre de ligne
	 */
	public int getLargeur()
	{
		return this.m_l;
	}
	
	/**
	 * Retourner le nombre de colonne de la grille.
	 * 
	 * @return Nombre de colonne
	 */
	public int getLongueur()
	{
		return this.m_c;
	}
	
	/**
	 * Retourner une cellule en fonction de ces coordoones (l;c)
	 * 
	 * @param l
	 * @param c
	 * @return Une cellule
	 */
	public Cellule getCellule(int l, int c)
	{
		return this.m_grille[l][c];
	}
	
	/**
	 * Retourne la cellule du dernier coup jou.
	 * 
	 * @return Une cellule.
	 */
	public Cellule getDernierCoup()
	{
		return this.m_journal.get(0);
	}
	
	/**
	 * Retourne le nombre d'alignement pour un joueur.
	 * 
	 * @param symbole
	 * @return Nombre d'alignement.
	 */
	public int getScore(char symbole)
	{
		if (symbole == 'X')
			return this.m_alignX.size();
		
		return this.m_alignO.size();
	}
	
	/**
	 * Appliquer un symbole sur une cellule.
	 * A chaque application on met  jour le journal des coups.
	 * On dtecte un alignement en fonction du dernier coups jous et de ses voisines.
	 * 
	 * @param symbole
	 * @param l
	 * @param c
	 * @throws Throwable
	 */
	public void appliquer(char symbole, int l, int c) throws Throwable
	{
		if (symbole != 'X' && symbole != 'O')
			throw new Throwable("-2.1");
		
		if (l < 0 || l >= this.m_l)
			throw new Throwable("-2.2");
		
		if (c < 0 || c >= this.m_c)
			throw new Throwable("-2.3");
		
		if (this.m_grille[l][c] != null)
			throw new Throwable("-3.1");
		
		this.m_grille[l][c] = new Cellule(this, symbole, l, c);
		this.m_journal.add(0, this.m_grille[l][c]);
		
		ArrayList<String> voisines = this.getCellule(l, c).getVoisines();
		
		if (voisines.contains("N") || voisines.contains("S")) {
			this.alignementV();
			
			if (voisines.contains("N") && voisines.contains("S"))
				this.alignementV();
		}
		
		if (voisines.contains("E") || voisines.contains("O")) {
			this.alignementH();
			
			if (voisines.contains("E") && voisines.contains("O"))
				this.alignementH();
		}
		
		if (voisines.contains("NE") || voisines.contains("SO")) {
			this.alignementD2();
			
			if (voisines.contains("NE") && voisines.contains("SO"))
				this.alignementD2();
		}
		
		if (voisines.contains("SE") || voisines.contains("NO")) {
			this.alignementD1();
			
			if (voisines.contains("SE") && voisines.contains("NO"))
				this.alignementD1();
		}
		
	}
	
	/**
	 * Calculer un alignement vertical en fonction du dernier coup jou.
	 */
	public void alignementV()
	{
		int compteur  = 1;
		Cellule debut = this.getDernierCoup();
		Cellule fin   = debut;
		
		while(compteur < 5 && debut.getVoisines().contains("N") && !debut.estExtremite('|')) {
			compteur++;
			debut = this.getCellule(debut.getLigne() - 1, debut.getColonne());			
		}
		
		while(compteur < 5 && fin.getVoisines().contains("S") && (compteur == 1 || !fin.estExtremite('|'))) {
			compteur++;
			fin = this.getCellule(fin.getLigne() + 1, fin.getColonne());			
		}
			
		if (compteur == 5) {
			Alignement align = new Alignement('|', debut, fin);
			
			if (this.getDernierCoup().getSymbole() == 'X')
				this.m_alignX.add(align);
			else
				this.m_alignO.add(align);
			
			for (int i = 0; i < 5; i++)
				this.getCellule(debut.getLigne() + i, debut.getColonne()).addAlign(align);
		}
	}
	
	/**
	 * Calculer un alignement horizontal en fonction du dernier coup jou.
	 */
	public void alignementH()
	{
		int compteur  = 1;
		Cellule debut = this.getDernierCoup();
		Cellule fin   = debut;
		
		while(compteur < 5 && debut.getVoisines().contains("O") && !debut.estExtremite('-')) {
			compteur++;
			debut = this.getCellule(debut.getLigne(), debut.getColonne() - 1);			
		}
		
		while(compteur < 5 && fin.getVoisines().contains("E") && (compteur == 1 || !fin.estExtremite('-'))) {
			compteur++;
			fin = this.getCellule(fin.getLigne(), fin.getColonne() + 1);			
		}
			
		if (compteur == 5) {
			Alignement align = new Alignement('|', debut, fin);
			
			if (this.getDernierCoup().getSymbole() == 'X')
				this.m_alignX.add(align);
			else
				this.m_alignO.add(align);
			
			for (int i = 0; i < 5; i++)
				this.getCellule(debut.getLigne(), debut.getColonne() + i).addAlign(align);
		}
	}
	
	/**
	 * Calculer un alignement du NO au SE en fonction du dernier coup jou.
	 */
	public void alignementD1()
	{
		int compteur  = 1;
		Cellule debut = this.getDernierCoup();
		Cellule fin   = debut;
		
		while(compteur < 5 && debut.getVoisines().contains("NO") && !debut.estExtremite('\\')) {
			compteur++;
			debut = this.getCellule(debut.getLigne() - 1, debut.getColonne() - 1);
		}
		
		while(compteur < 5 && fin.getVoisines().contains("SE") && (compteur == 1 || !fin.estExtremite('\\'))) {
			compteur++;
			fin = this.getCellule(fin.getLigne() + 1, fin.getColonne() + 1);			
		}
			
		if (compteur == 5) {
			Alignement align = new Alignement('\\', debut, fin);
			
			if (this.getDernierCoup().getSymbole() == 'X')
				this.m_alignX.add(align);
			else
				this.m_alignO.add(align);
			
			for (int i = 0; i < 5; i++)
				this.getCellule(debut.getLigne() + i, debut.getColonne() + i).addAlign(align);
		}
	}
	
	/**
	 * Calculer un alignement du NE au SO en fonction du dernier coup jou.
	 */
	public void alignementD2()
	{
		int compteur  = 1;
		Cellule debut = this.getDernierCoup();
		Cellule fin   = debut;
		
		while(compteur < 5 && debut.getVoisines().contains("NE") && !debut.estExtremite('/')) {
			compteur++;
			debut = this.getCellule(debut.getLigne() - 1, debut.getColonne() + 1);			
		}
		
		while(compteur < 5 && fin.getVoisines().contains("SO") && (compteur == 1 || !fin.estExtremite('/'))) {
			compteur++;
			fin = this.getCellule(fin.getLigne() + 1, fin.getColonne() - 1);			
		}
			
		if (compteur == 5) {
			Alignement align = new Alignement('/', debut, fin);
			
			if (this.getDernierCoup().getSymbole() == 'X')
				this.m_alignX.add(align);
			else
				this.m_alignO.add(align);
			
			for (int i = 0; i < 5; i++)
				this.getCellule(debut.getLigne() + i, debut.getColonne() - i).addAlign(align);
		}
	}
	
	/**
	 * @return L'affichage de l'objet
	 */
	public String toString()
	{
		String ch = new String("");
		
		for (int i = 0; i < this.m_l; i++) {   
			for(int j = 0; j < this.m_c; j++)
				if (this.getCellule(i, j) != null)
					ch += " "+ this.m_grille[i][j].toString() +" ";
				else
					ch += " ------ ";
			
			ch += "\n";
		}
		
		return ch;
	}
	
	/**
	 * Grille.Cellule
	 * 
	 * @package 	_Morpion
	 * @author  	KOVAC Alexandre
	 * @version 	1.0.0
	 * @link 		<http://www.kovacou.fr/University/Project/Morpion>
	 */
	public class Cellule
	{
		private int m_l;
		private int m_c;
		private char m_symbole;
		private Grille m_grille;
		private ArrayList<Alignement> m_align;
		
		/** ___________
		 * |1xxxx2xxxx3|
		 * |4	 9    5|
		 * |6xxxx7xxxx8|
		 */
		private int m_position;
		
		/**
		 * Constructeur.
		 * Construit une nouvelle cellule. 
		 * 
		 * @param grille La grille de jeu
		 * @param symbole Symbole (X ou O)
		 * @param l Numro de la ligne
		 * @param c Numro de la colonne
		 */
		public Cellule(Grille grille, char symbole, int l, int c)
		{
			this.m_l       = l;
			this.m_c       = c;
			this.m_symbole = symbole;
			this.m_grille  = grille;
			this.m_align   = new ArrayList<Alignement>();
			
			// Recherche des positions particulires
			if (this.m_l - 1 < 0) {
				if (this.m_c - 1 < 0)
					m_position = 1;
				else if (this.m_c + 1 >= this.m_grille.getLongueur())
					m_position = 3;
				else
					m_position = 2;
			}
			else if (this.m_l + 1 == this.m_grille.getLargeur()) {
				if (this.m_c - 1 < 0)
					m_position = 6;
				else if (this.m_c + 1 == this.m_grille.getLongueur())
					m_position = 8;
				else
					m_position = 7;
			}
			else {
				if (this.m_c - 1 < 0)
					m_position = 4;
				else if (this.m_c + 1 == this.m_grille.getLongueur())
					m_position = 5;
				else
					m_position = 9;
			}
		}
		
		/**
		 * Retourner les voisines de la cellule courante.
		 * Exemple : [N, S], [NE, SO]
		 * 
		 * @return ArrayList
		 */
		public ArrayList<String> getVoisines()
		{
			ArrayList<String> arrayVoisines = new ArrayList<String>();			
			
			// Voisine S
			if ((m_position == 1 || m_position == 2 || m_position == 3 || m_position == 4 || m_position == 9 || 
				 m_position == 5) && this.m_grille.getCellule(this.m_l + 1, this.m_c) != null && 
				 this.m_grille.getCellule(this.m_l + 1, this.m_c).getSymbole() == this.getSymbole())
				arrayVoisines.add("S");
			
			// Voisine N
			if ((m_position == 7 || m_position == 8 || m_position == 6 || m_position == 4 || m_position == 9 || 
			     m_position == 5) && this.m_grille.getCellule(this.m_l - 1, this.m_c) != null &&
			     this.m_grille.getCellule(this.m_l - 1, this.m_c).getSymbole() == this.getSymbole())
					arrayVoisines.add("N");
			
			// Voisine E
			if ((m_position == 1 || m_position == 2 || m_position == 6 || m_position == 4 || m_position == 9 || 
				 m_position == 7) && this.m_grille.getCellule(this.m_l, this.m_c + 1) != null 
				 && this.m_grille.getCellule(this.m_l, this.m_c + 1).getSymbole() == this.getSymbole())
					arrayVoisines.add("E");
			
			// Voisine O
			if ((m_position == 7 || m_position == 2 || m_position == 3 || m_position == 8 || m_position == 9 || 
				 m_position == 5) && this.m_grille.getCellule(this.m_l, this.m_c - 1) != null && 
				 this.m_grille.getCellule(this.m_l, this.m_c - 1).getSymbole() == this.getSymbole())
					arrayVoisines.add("O");
			
			// Voisine NE
			if ((m_position == 4 || m_position == 9 || m_position == 6 || m_position == 7) && 
				 this.m_grille.getCellule(this.m_l - 1, this.m_c + 1) != null &&
				 this.m_grille.getCellule(this.m_l - 1, this.m_c + 1).getSymbole() == this.getSymbole()) {
					arrayVoisines.add("NE");
			}
			
			// Voisine NO
			if ((m_position == 5 || m_position == 9 || m_position == 8 || m_position == 7) && 
				 this.m_grille.getCellule(this.m_l - 1, this.m_c - 1) != null &&
				 this.m_grille.getCellule(this.m_l - 1, this.m_c - 1).getSymbole() == this.getSymbole())
					arrayVoisines.add("NO");
			
			// Voisine SE
			if ((m_position == 4 || m_position == 9 || m_position == 1 || m_position == 2) && 
				 this.m_grille.getCellule(this.m_l + 1, this.m_c + 1) != null && 
			     this.m_grille.getCellule(this.m_l + 1, this.m_c + 1).getSymbole() == this.getSymbole())
					arrayVoisines.add("SE");
			
			// Voisine SO
			if ((m_position == 9 || m_position == 5 || m_position == 2 || m_position == 3) && 
				 this.m_grille.getCellule(this.m_l + 1, this.m_c - 1) != null &&
			     this.m_grille.getCellule(this.m_l + 1, this.m_c - 1).getSymbole() == this.getSymbole())
					arrayVoisines.add("SO");
			
			return arrayVoisines;
		}
		
		/**
		 * Ajouter un nouvel alignement dans lequel se trouve cette cellule.
		 * 
		 * @param align
		 */
		public void addAlign(Alignement align)
		{
			this.m_align.add(align);
		}
		
		/**
		 * Retourner la ligne de la cellule.
		 * 
		 * @return La ligne
		 */
		public int getLigne()
		{
			return this.m_l;
		}
		
		/**
		 * Retourner la colonne de la cellule.
		 * 
		 * @return La colonne
		 */
		public int getColonne()
		{
			return this.m_c;
		}
		
		/**
		 * Retourner le symbole de la cellule.
		 * 
		 * @return Un symbole
		 */
		public char getSymbole()
		{
			return this.m_symbole;
		}
		
		/**
		 * Retourner la position (ou zone) de la cellule.
		 * 
		 * @return La position
		 */
		public int getPosition()
		{
			return m_position;
		}
		
		/**
		 * Dterminer si la cellule courante est une extrmit d'un alignement.
		 * 
		 * @param orientation
		 * @return boolean
		 */
		public boolean estExtremite(char orientation)
		{
			if (this.m_align.size() == 0)
				return false;
			
			for (int i = 0; i < this.m_align.size(); i++)				
				if (this.m_align.get(i).getOrientation() == orientation && (this.equals(this.m_align.get(i).getDebut()) || this.equals(this.m_align.get(i).getFin())))
					return true;
			
			return false;
		}
		
		/**
		 * @return L'affichage de l'objet
		 */
		public String toString()
		{
			return this.m_symbole +"("+ this.m_l +";"+ this.m_c +")";
		}
		
		/**
		 * Compare l'objet support avec un objet extrieur.
		 * Comparaison par contenu.
		 * 
		 * @param c
		 * @return Boolean
		 */
		public boolean equals(Cellule c)
		{
			return c.getColonne() == this.m_c && c.getLigne() == this.m_l;
		}
	}
	
	/**
	 * Grille.Alignement
	 * 
	 * @package 	_Morpion
	 * @author  	KOVAC Alexandre
	 * @version 	1.0.0
	 * @link 		<http://www.kovacou.fr/University/Project/Morpion>
	 */
	public class Alignement
	{
		private Cellule m_debut;
		private Cellule m_fin;
		private char m_orientation; // (/\-|)
		
		/**
		 * Constructeur.
		 * Construit un nouvel alignement.
		 * 
		 * @param orientation
		 * @param debut
		 * @param fin
		 */
		public Alignement(char orientation, Cellule debut, Cellule fin)
		{
			this.m_orientation = orientation;
			this.m_debut       = debut;
			this.m_fin         = fin;
		}
		
		/**
		 * Retourner la cellule qui dbut l'alignement.
		 * 
		 * @return Cellule du dbut de l'alignement
		 */
		public Cellule getDebut()
		{
			return this.m_debut;
		}
		
		/**
		 * Retourner la cellule qui finie l'alignement.
		 * 
		 * @return Cellule de fin de l'alignement
		 */
		public Cellule getFin()
		{
			return this.m_fin;
		}
		
		/**
		 * Retourner l'orientation de l'alignement.
		 * 
		 * @return Orientation de l'alignement
		 */
		public char getOrientation()
		{
			return this.m_orientation;
		}
		
		/**
		 * Vrifie si une cellule appartient  cet alignement.
		 * 
		 * @return Boolean
		 */
		public boolean appartient(Cellule c)
		{			
			if (this.getOrientation() == '-') {
				if (c.getLigne() == m_debut.getLigne()) {
					if (m_debut.getColonne() - m_fin.getColonne() < 0)
						return c.getColonne() >= m_debut.getColonne() && c.getColonne() <= m_fin.getColonne();
					else
		                return c.getColonne() <= m_debut.getColonne() && c.getColonne() >= m_fin.getColonne();   
				}
			}
			else if(this.getOrientation() == '|') {
				if (c.getColonne() == m_debut.getColonne()) {
					if (m_debut.getLigne() - m_fin.getLigne() < 0)
						return c.getLigne() >= m_debut.getLigne() && c.getLigne() <= m_fin.getLigne();
					else
						return c.getLigne() <= m_debut.getLigne() && c.getLigne() >= m_fin.getLigne();
			   }
			}
			else if (this.getOrientation() == '\\') {
			   if (m_debut.getColonne() - m_fin.getColonne() < 0) {
				   int i = m_debut.getLigne();
				   int j = m_debut.getColonne();
		     	 
				   for (int k = 0; k < 5; k++) {
					   if (c.getColonne() == j && c.getLigne() == i)
			                 return true;
					   
					   i++;
					   j++;
				   }
			   }
			   else {
				   int i = m_debut.getLigne();
				   int j = m_debut.getColonne();
		     	 
				   for (int k = 0; k < 5; k++) {
					   if (c.getColonne() == j && c.getLigne() == i)
			                 return true;
					   
					   i--;
					   j--;
				   }
			   }	
		   }
		  else if (this.getOrientation() == '/') {
			  if (m_debut.getColonne() - m_fin.getColonne() < 0) {
				   int i = m_debut.getLigne();
				   int j = m_debut.getColonne();
		     	 
				   for (int k = 0; k < 5; k++) {
					   if (c.getColonne() == j && c.getLigne() == i)
			                 return true;
					   
					   i++;
					   j--;
				   }
			   }
			   else {
				   int i = m_debut.getLigne();
				   int j = m_debut.getColonne();
		     	 
				   for (int k = 0; k < 5; k++) {
					   if (c.getColonne() == j && c.getLigne() == i)
			                 return true;
					   
					   i--;
					   j++;
				   }
			   	}	
		  	}
			
			return false;
		}
		
		
		/**
		 * Affichage de l'objet.
		 */
		public String toString()
		{
			return this.m_orientation + "["+ this.m_debut.toString() + this.m_fin.toString() +"]";
		}
	}
}
